local XPerl_Party_Events = {}
local checkRaidNextUpdate
local PartyFrames = {}

----------------------
-- Loading Function --
----------------------
function XPerl_Party_Events_OnLoad(self)
	local events = {"PLAYER_ENTERING_WORLD", "PARTY_MEMBER_ENABLE", "PARTY_MEMBER_DISABLE", "RAID_ROSTER_UPDATED", "PARTY_MEMBERS_CHANGED",
			"UNIT_COMBAT", "UNIT_FACTION", "UNIT_DYNAMIC_FLAGS", "UNIT_FLAGS", "UNIT_AURA", "UNIT_MODEL_CHANGED",
			"UNIT_TARGET", "UNIT_RAGE", "UNIT_MAXRAGE", "UNIT_ENERGY", "UNIT_MAXENERGY", "UNIT_MANA", "UNIT_MAXMANA",
			"UNIT_HEALTH", "UNIT_MAXHEALTH", "UNIT_LEVEL", "UNIT_DISPLAYPOWER", "UNIT_NAME_UPDATE", "PLAYER_FLAGS_CHANGED"}
	for i,event in pairs(events) do
		self:RegisterEvent(event)
	end

	XPerl_BlizzFrameDisable(PartyMemberFrame1)
	XPerl_BlizzFrameDisable(PartyMemberFrame2)
	XPerl_BlizzFrameDisable(PartyMemberFrame3)
	XPerl_BlizzFrameDisable(PartyMemberFrame4)

	hooksecurefunc("ShowPartyFrame",
		function()
			if (not InCombatLockdown()) then
				for i = 1,4 do
					getglobal("PartyMemberFrame"..i):Hide()
				end
			end
		end
	)

	self:SetScript("OnEvent", XPerl_Party_OnEvent)

	XPerl_RegisterOptionChanger(XPerl_Party_Set_Bits, "Party")
end

-- SetPartyFrame
local function SetPartyFrame(value, self)
	for k,v in pairs(PartyFrames) do
		if (v == self) then
			PartyFrames[k] = nil
		end
	end

	PartyFrames[value] = self
end

-- onAttrChanged
local function onAttrChanged(self, name, value)
	if (name == "unit") then
		if (value and value ~= "party0") then
			SetPartyFrame(value, self)
			XPerl_Party_UpdateDisplay(self)
			if (XPerl_ArcaneBar_RegisterFrame) then
				XPerl_ArcaneBar_RegisterFrame(self.nameFrame, value, "ArcaneBarParty")
			end
		end
	end
end

-- XPerl_Party_OnLoad
function XPerl_Party_OnLoad(self)
	XPerl_BaseFrameProperties(self)

	XPerl_Party_SecureHeader:SetAttribute("child"..self:GetID(), self)

	self.time = 0

	--self.highlight.tex:SetVertexColor(0.86, 0.82, 0.41)
	XPerl_RegisterHighlight(self.highlight, 3)
	XPerl_RegisterHighlight(self.targetFrame, 3)
	XPerl_RegisterPerlFrames(self, {"NameFrame", "StatsFrame", "PortraitFrame", "LevelFrame", "TargetFrame"})

	self.FlashFrames = {self.nameFrame, self.levelFrame, self.statsFrame, self.portraitFrame}

	self:SetScript("OnUpdate", XPerl_Party_OnUpdate)
	self:SetScript("OnAttributeChanged", onAttrChanged)
end

-- ShowHideValues
local function ShowHideValues(self)
	if (XPerlConfig.ShowPartyValues == 1) then
        	self.statsFrame.healthBar.text:Show()
        	self.statsFrame.manaBar.text:Show()
	else
        	self.statsFrame.healthBar.text:Hide()
        	self.statsFrame.manaBar.text:Hide()
	end
end

-- XPerl_Party_UpdateManaType
local function XPerl_Party_UpdateManaType(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		XPerl_SetManaBarType(partyid, self.statsFrame.manaBar)
	end
end

-- XPerl_Party_UpdateHealth
local function XPerl_Party_UpdateHealth(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		local numID = strmatch(partyid, "party(%d)")			-- TODO - will be wrong later
		local Partyhealth = UnitHealth(partyid)
		local Partyhealthmax = UnitHealthMax(partyid)
		local reason

		if (Partyhealth > Partyhealthmax) then
			-- New glitch with 1.12.1
			if (UnitIsDeadOrGhost(partyid)) then
				Partyhealth = 0
			else
				Partyhealth = Partyhealthmax
			end
		end

		if (self.feigning and not XPerl_IsFeignDeath(partyid)) then
			self.feigning = nil
		end
		if (self.feigning) then
			self.statsFrame.healthBar:SetMinMaxValues(0, 1)
		        self.statsFrame.healthBar:SetValue(0)
			XPerl_SetSmoothBarColor(self.statsFrame.healthBar, 0)
		else
			XPerl_SetHealthBar(self.statsFrame.healthBar, Partyhealth, Partyhealthmax, XPerlConfig.HealerModeParty, XPerlConfig.HealerModePartyType)
		end

		self.statsFrame.healthBar.percent:SetText(string.format("%d",(100*(Partyhealth / Partyhealthmax))+0.5).."%")

		if (not UnitIsConnected(partyid)) then
			hide = XPERL_LOC_OFFLINE
		else
			if (self.feigning) then
				reason = XPERL_LOC_FEIGNDEATH

			elseif (self.afk) then
				reason = CHAT_MSG_AFK

			elseif (self.dnd) then
				reason = CHAT_MSG_DND

			elseif (UnitIsDead(partyid)) then
				reason = XPERL_LOC_DEAD

			elseif (UnitIsGhost(partyid)) then
				reason = XPERL_LOC_GHOST

			elseif ((Partyhealth==1) and (Partyhealthmax==1)) then
				reason = XPERL_LOC_UPDATING
			end
		end

		ShowHideValues(self)
		if (reason) then
			self.statsFrame.healthBar.text:SetText(reason)
			self.statsFrame.healthBar.text:Show()
			self.statsFrame.healthBar:SetStatusBarColor(0.5, 0.5, 0.5, 1)
			self.statsFrame.healthBar.bg:SetVertexColor(0.5, 0.5, 0.5, 0.5)
			self.statsFrame.manaBar:SetStatusBarColor(0.5, 0.5, 0.5, 1)
			self.statsFrame.manaBar.bg:SetVertexColor(0.5, 0.5, 0.5, 0.5)
			self.greyMana = true
		else
			if (self.greyMana) then
				self.greyMana = nil
				XPerl_Party_UpdateManaType(self)
			end
		end
	end
end

-- XPerl_Party_UpdatePlayerFlags(self)
local function XPerl_Party_UpdatePlayerFlags(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		if (UnitIsAFK(partyid)) then
			if (not self.afk) then
				change = true
				self.afk = GetTime()
				self.dnd = nil
			end
		elseif (UnitIsDND(partyid)) then
			if (not self.dnd) then
				change = true
				self.dnd = GetTime()
				self.afk = nil
			end
		else
			if (self.afk or self.dnd) then
				self.afk, self.dnd = nil, nil
				change = true
			end
		end

		if (change) then
			XPerl_Party_UpdateHealth(self)
		end
	end
end

--------------------
-- Buff Functions --
--------------------

-- GetBuffButton
-- debuff must be 1 or 0, as it's used in size calc
local function GetBuffButton(self, buffnum, debuff, createIfAbsent)

	local buffType
	if (debuff == 1) then
		buffType = "DeBuff"
	else
		buffType = "Buff"
	end
	local name = self:GetName().."_BuffFrame_"..buffType..buffnum
	local button = getglobal(name)

	if (not button and createIfAbsent) then
		button = CreateFrame("Button", name, self.buffFrame, "XPerl_"..buffType.."Template")
		button:SetID(buffnum)
		button:SetAlpha(self:GetAlpha())		-- New frames always seem to have 100% alpha

		local size = XPerlConfig.PartyBuffSize * (1 + (0.4 * debuff))
		button:SetHeight(size)
		button:SetWidth(size)

		if (debuff == 1) then
			button:SetScript("OnEnter", XPerl_Party_SetDeBuffTooltip)
		else
			button:SetScript("OnEnter", XPerl_Party_SetBuffTooltip)
		end
		button:SetScript("OnLeave", XPerl_PlayerTipHide)

		button:ClearAllPoints()
		if (buffnum == 1) then
			if (debuff == 1) then
				button:SetPoint("LEFT", self.statsFrame, "RIGHT", 0, 0)
			else
				button:SetPoint("TOPLEFT", self.buffFrame, "TOPLEFT", 0, 0)
			end
		else
			local prevButton = getglobal(self:GetName().."_BuffFrame_"..buffType..(buffnum - 1))
			button:SetPoint("TOPLEFT", prevButton, "TOPRIGHT", 1 + debuff, 0)
		end
	end

	return button
end

-- XPerl_Party_SetDebuffLocation
function XPerl_Party_SetDebuffLocation(self)
	local debuff1 = getglobal(self:GetName().."_BuffFrame_DeBuff1")
	if (debuff1) then
        	debuff1:ClearAllPoints()

		if (XPerlConfig.PartyDebuffsBelow == 0) then
			local petFrame = getglobal("XPerl_partypet"..self:GetID())
			if (petFrame and petFrame:IsVisible()) then
        	                debuff1:SetPoint("TOPLEFT", petFrame.nameFrame, "TOPRIGHT", 0, -4)
			else
        	                debuff1:SetPoint("TOPLEFT", self.statsFrame, "TOPRIGHT", 0, -4)
			end
		else
			local buff1 = getglobal(self:GetName().."_BuffFrame_Buff1")
			if (not buff1) then
        	        	debuff1:SetPoint("TOPLEFT", self.buffFrame, "TOPLEFT", 0, -20)
			else
        	        	debuff1:SetPoint("TOPLEFT", buff1, "BOTTOMLEFT", 0, -2)
			end
		end
	end
end

-- XPerl_Party_SetBuffLocation
local function XPerl_Party_SetBuffLocation(self)
	local buff1 = getglobal(self:GetName().."_BuffFrame_Buff1")
	if (buff1) then
        	buff1:ClearAllPoints()
		buff1:SetPoint("TOPLEFT", self.buffFrame, "TOPLEFT", 0, 0)
	end
end

-- XPerl_Party_Buff_UpdateAll
local function XPerl_Party_Buff_UpdateAll(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		local thisid = self:GetName()

		if (XPerlConfig.PartyBuffs == 0 and XPerlConfig.PartyDebuffs == 0) then
			self.buffFrame:Hide()
		else
			self.buffFrame:Show()

			if (UnitExists(partyid)) then
				if (XPerlConfig.PartyBuffs == 1) then
					for buffnum=1,16 do
						local name, rank, buff, count = XPerl_UnitBuff(partyid, buffnum, XPerlConfig.PartyCastableBuffs)

						if (buff) then
							local button = GetBuffButton(self, buffnum, 0, true)
							button.icon:SetTexture(buff)
							if (count > 1) then
								button.count:SetText(count)
								button.count:Show()
							else
								button.count:Hide()
							end
							button:Show()
						else
							local button = getglobal(thisid.."_BuffFrame_Buff"..buffnum)
							if (button) then
								button:Hide()
							end
						end
					end

					XPerl_Party_SetBuffLocation(self)
				else
					for buffnum=1,16 do
						local button = getglobal(thisid.."_BuffFrame_Buff"..buffnum)
						if (button) then
							button:Hide()
						end
					end
				end

				if (XPerlConfig.PartyDebuffs == 1) then
					for buffnum=1,8 do
						local name, rank, buff, count, debuffType = XPerl_UnitDebuff(partyid, buffnum, XPerlConfig.PartyCurableDebuffs)

						if (buff) then
							local button = GetBuffButton(self, buffnum, 1, buff)

							button.icon:SetTexture(buff)

							if (count > 1) then
								button.count:SetText(count)
								button.count:Show()
							else
								button.count:Hide()
							end

							local borderColor = DebuffTypeColor[(debuffType or "none")]
							local buffBorder = getglobal(button:GetName().."Border")
							buffBorder:SetVertexColor(borderColor.r, borderColor.g, borderColor.b)

							button:Show()
						else
							local button = getglobal(thisid.."_BuffFrame_DeBuff"..buffnum)
							if (button) then
								button:Hide()
							end
						end
					end

					XPerl_Party_SetDebuffLocation(self)
				else
					for buffnum=1,8 do
						local button = getglobal(thisid.."_BuffFrame_DeBuff"..buffnum)
						if (button) then
							button:Hide()
						end
					end
				end
			end
		end

		local _, class = UnitClass(partyid)
		if (class == "HUNTER") then
			local feigning = XPerl_IsFeignDeath(partyid)
			if (feigning ~= self.feigning) then
				self.feigning = feigning
				XPerl_Party_UpdateHealth(self)
			end
		end

		XPerl_CheckDebuffs(self)
	end
end

-- XPerl_Party_SetBuffTooltip
function XPerl_Party_SetBuffTooltip(self)
	local partyid = self:GetParent():GetParent():GetAttribute("unit")
	if (partyid) then
		GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT",30,0)
		XPerl_TooltipSetUnitBuff(GameTooltip, partyid, self:GetID(), XPerlConfig.PartyCastableBuffs)
		XPerl_ToolTip_AddBuffDuration(partyid)
	end
end

-- XPerl_Party_SetDeBuffTooltip
function XPerl_Party_SetDeBuffTooltip(self)
	local partyid = self:GetParent():GetParent():GetAttribute("unit")
	if (partyid) then
		GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT",30,0)
		XPerl_TooltipSetUnitDebuff(GameTooltip, partyid, self:GetID(), XPerlConfig.PartyCurableDebuffs)
		XPerl_ToolTip_AddDebuffDuration(partyid, nil, self:GetID(), XPerlConfig.PartyCurableDebuffs)
	end
end

-------------------------
-- The Update Function --
-------------------------
local function XPerl_Party_CombatFlash(self, elapsed, argNew, argGreen)
	if (XPerl_CombatFlashSet (elapsed, self, argNew, argGreen)) then
		XPerl_CombatFlashSetFrames(self)
	end
end

-- XPerl_Party_UpdatePortrait
function XPerl_Party_UpdatePortrait(self)
	if (XPerlConfig.ShowPartyPortrait == 1) then
		local partyid = self:GetAttribute("unit")
		if (partyid) then
			if (XPerlConfig.ShowPartyPortrait3D == 1 and UnitIsVisible(partyid)) then
				self.portraitFrame.portrait:Hide()
				self.portraitFrame.portrait3D:Show()
				XPerlSetPortrait3D(self.portraitFrame.portrait3D, partyid)
			else
				self.portraitFrame.portrait3D.last3DTime = nil
				self.portraitFrame.portrait:Show()
				self.portraitFrame.portrait3D:Hide()
				SetPortraitTexture(self.portraitFrame.portrait, partyid)
			end
		end
	end
end

-- XPerl_Party_UpdateName
local function XPerl_Party_UpdateName(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		local Partyname = UnitName(partyid)
		if (Partyname) then
			self.nameFrame.name:SetFontObject(GameFontNormal)
        	        self.nameFrame.name:SetText(Partyname)

			if (self.nameFrame.name:GetStringWidth() > self.nameFrame:GetWidth() - 4) then
				self.nameFrame.name:SetFontObject(GameFontNormalSmall)
			else
				self.nameFrame.name:SetFontObject(GameFontNormal)
			end

			XPerl_ColourFriendlyUnit(self.nameFrame.name, partyid)
		end
	end
end

-- XPerl_Party_UpdateLeader
local function XPerl_Party_UpdateLeader(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		if ("party"..GetPartyLeaderIndex() == partyid) then
			self.nameFrame.leaderIcon:Show()
		else
			self.nameFrame.leaderIcon:Hide()
		end

		local lootMethod
		local lootMaster
		lootMethod, lootMaster = GetLootMethod()
		if ( partyid == "party"..(lootMaster or "nil") ) then
			self.nameFrame.masterIcon:Show()
		else
			self.nameFrame.masterIcon:Hide()
		end
	end
end

-- XPerl_Party_Update_Control
local function XPerl_Party_Update_Control(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		if (UnitIsVisible(partyid) and UnitIsCharmed(partyid)) then
			self.nameFrame.warningIcon:Show()
		else
			self.nameFrame.warningIcon:Hide()
		end
	end
end

-- XPerl_Party_UpdatePVP
local function XPerl_Party_UpdatePVP(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		if (XPerlConfig.ShowPartyPVP == 1 and UnitIsPVP(partyid)) then
			self.nameFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..(UnitFactionGroup(partyid) or "FFA"))
			self.nameFrame.pvpIcon:Show()
		else
			self.nameFrame.pvpIcon:Hide()
		end
	end
end

-- XPerl_Party_UpdateCombat
function XPerl_Party_UpdateCombat(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		XPerl_Party_UpdateCombat1(self, partyid)
		XPerl_Party_Update_Control(self)
	end
end

-- XPerl_Party_UpdateCombat1
function XPerl_Party_UpdateCombat1(self, argID)
	if (UnitAffectingCombat(argID)) then
		self.nameFrame.combatIcon:Show()
	else
		self.nameFrame.combatIcon:Hide()
	end
end

-- XPerl_Party_UpdateLevel
local function XPerl_Party_UpdateLevel(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		local Partylevel = UnitLevel(partyid)
		local color = GetDifficultyColor(Partylevel)

		self.levelFrame.text:SetTextColor(color.r,color.g,color.b)
		self.levelFrame.text:SetText(Partylevel)
	end
end

-- XPerl_Party_UpdateClass
local function XPerl_Party_UpdateClass(self)
	local partyid = self:GetAttribute("unit")
	if (partyid and UnitIsPlayer(partyid)) then
		local _, PlayerClass = UnitClass(partyid)
		local r, l, t, b = XPerl_ClassPos(PlayerClass)
		self.levelFrame.classTexture:SetTexCoord(r, l, t, b)
	end

	if (XPerlConfig.ShowPartyClassIcon == 1) then
	        self.levelFrame.classTexture:Show()
	else
	        self.levelFrame.classTexture:Hide()
	end
end

-- XPerl_Party_UpdateMana
local function XPerl_Party_UpdateMana(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		local Partymana = UnitMana(partyid)
		local Partymanamax = UnitManaMax(partyid)

		if (Partymanamax == 1 and Partymana > Partymanamax) then
			Partymanamax = Partymana
		end

		self.statsFrame.manaBar:SetMinMaxValues(0, Partymanamax)
		self.statsFrame.manaBar:SetValue(Partymana)
		pmanaPct = (Partymana * 100.0) / Partymanamax
		pmanaPct =  string.format("%3.0f", pmanaPct)

		if (UnitPowerType(partyid)>=1) then
		        self.statsFrame.manaBar.percent:SetText(Partymana)
		else
		        self.statsFrame.manaBar.percent:SetText(string.format("%d",(100*(Partymana / Partymanamax))+0.5).."%")
		end

		if (XPerlConfig.ShowPartyValues == 1) then
			self.statsFrame.manaBar.text:Show()
		else
			self.statsFrame.manaBar.text:Hide()
		end

		self.statsFrame.manaBar.text:SetText(Partymana.."/"..Partymanamax)

		local numID = strmatch(partyid, "party(%d)")			-- TODO - will be wrong later
		local petFrame = getglobal("XPerl_partypet"..numID)

		if (not UnitIsConnected(partyid)) then
		        self.statsFrame.healthBar.text:SetText(XPERL_LOC_OFFLINE)
		end
	end
end

-- XPerl_Party_UpdateRange
local function XPerl_Party_UpdateRange(self)
	local partyid = self:GetAttribute("unit")
	if (partyid) then
		if (XPerlConfig.ShowParty30YardSymbol == 0 or CheckInteractDistance(partyid, 4) or not UnitIsConnected(partyid)) then
			self.nameFrame.rangeIcon:Hide()
		else
			self.nameFrame.rangeIcon:Show()
			self.nameFrame.rangeIcon:SetAlpha(1)
		end
	end
end

-- CheckRaid
local function CheckRaid()
	if (InCombatLockdown()) then
		tinsert(XPerl_OutOfCombatQueue, CheckRaid)
	else
		XPerl_Party_Anchor:StopMovingOrSizing()
		if (UnitInRaid("player")) then
			if (XPerlConfig.ShowPartyRaid == 1) then
				XPerl_Party_SecureHeader:Show()
			else
				XPerl_Party_SecureHeader:Hide()
			end
		else
			XPerl_Party_SecureHeader:Show()
		end
	end
end

-- XPerl_Party_OnUpdate
function XPerl_Party_OnUpdate(self, elapsed)
	if (self.PlayerFlash) then
		XPerl_Party_CombatFlash(self, arg1, false)
	end

	self.time = elapsed + self.time
	if (self.time >= 0.2) then
		self.time = 0

		XPerl_Party_UpdateRange(self)
		XPerl_UpdateSpellRange(self)
		XPerl_UpdateSpellRange(self.targetFrame)

		if (checkRaidNextUpdate) then
			checkRaidNextUpdate = checkRaidNextUpdate - 1
			if (checkRaidNextUpdate <= 0) then
				checkRaidNextUpdate = nil
				CheckRaid()
			end
		end
	end
end

-- XPerl_Party_UpdateDisplayAll
function XPerl_Party_UpdateDisplayAll()
	local any
	for i,frame in pairs(PartyFrames) do
		XPerl_Party_UpdateDisplay(frame)
		if (frame:IsShown()) then
			any = true
		end
	end
end

-- XPerl_Party_UpdateDisplay
function XPerl_Party_UpdateDisplay(self)
	local partyid = self:GetAttribute("unit")
	if (partyid and UnitExists(partyid)) then
		self.afk, self.dnd = nil,nil
		XPerl_Party_UpdateName(self)
		--XPerl_Party_UpdateLeader(self)		-- 2.0.4 testing need for this here.
		XPerl_Party_UpdateClass(self)
		XPerl_Party_UpdateManaType(self)
		XPerl_Party_UpdateHealth(self)
		XPerl_Party_UpdatePlayerFlags(self)
		XPerl_Party_UpdateMana(self)
		XPerl_Party_UpdateLevel(self)
		XPerl_Party_Update_Control(self)
        	XPerl_Party_UpdateCombat(self)
		XPerl_Party_UpdatePVP(self)
		XPerl_Party_UpdatePortrait(self)
		XPerl_Party_Buff_UpdateAll(self)
		XPerl_Party_UpdateTarget(self)
		XPerl_UpdateSpellRange(self)
	end
end

--------------------
-- Click Handlers --
--------------------

-- XPerl_Party_UpdateTarget
function XPerl_Party_UpdateTarget(self)
	if (XPerlConfig.ShowPartyTarget == 1) then
		local partyid = self:GetAttribute("unit")
		if (partyid) then
			if (UnitIsConnected(partyid) and UnitExists(partyid) and UnitIsVisible(partyid)) then
				local unit = partyid.."target"
				local targetname = UnitName(unit)

				if (targetname and targetname ~= UNKNOWNOBJECT) then
					self.targetFrame.name:SetText(targetname)

					XPerl_SetUnitNameColor (unit, self.targetFrame.name)

				        if (UnitAffectingCombat(unit)) then
				                self.targetFrame.combatIcon:SetTexCoord(0.5, 1.0, 0.0, 0.5)
				                self.targetFrame.combatIcon:Show()
					else
				                self.targetFrame.combatIcon:Hide()
					end

					if (XPerlConfig.ShowPartyPVP == 1 and UnitIsPVP(unit)) then
						self.targetFrame.pvpIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..(UnitFactionGroup(unit) or "FFA"))
						self.targetFrame.pvpIcon:Show()
					else
						self.targetFrame.pvpIcon:Hide()
					end
				end
			end
		end
	end
end

-------------------
-- Event Handler --
-------------------
function XPerl_Party_OnEvent(self, event, unit)
	local func = XPerl_Party_Events[event]
	if (func) then
		if (strmatch(event, "^UNIT_")) then
			local f = PartyFrames[unit]
			if (f) then
				func(f)
			end
		else
			func(self, unit)
		end
	else
XPerl_ShowMessage("EXTRA EVENT")
	end
end

-- PARTY_LEADER_CHANGED
function XPerl_Party_Events:PARTY_LEADER_CHANGED()
	for i,frame in pairs(PartyFrames) do
		XPerl_Party_UpdateLeader(frame)
	end
end

XPerl_Party_Events.PARTY_LOOT_METHOD_CHANGED	= XPerl_Party_Events.PARTY_LEADER_CHANGED

-- UNIT_COMBAT
function XPerl_Party_Events:UNIT_COMBAT()
	local partyid = self:GetAttribute("unit")
	if (arg1 == partyid) then
		XPerl_Party_UpdateCombat(self)
		if (arg2 == "HEAL") then
			XPerl_Party_CombatFlash(self, 0, true, true)
		elseif (arg4 and arg4 > 0) then
			XPerl_Party_CombatFlash(self, 0, true)
		end
	end
end

-- UNIT_HEALTH, UNIT_MAXHEALTH
function XPerl_Party_Events:UNIT_HEALTH()
	XPerl_Party_UpdateHealth(self)
end

-- PARTY_MEMBER_ENABLE
function XPerl_Party_Events:PARTY_MEMBER_ENABLE()
	for k,v in pairs(PartyFrames) do
		v.afk, v.dnd = nil,nil
		XPerl_Party_UpdateHealth(v)
		XPerl_Party_UpdatePlayerFlags(v)
		XPerl_Party_UpdateCombat(v)
		XPerl_Party_UpdateTarget(v)
	end
end

XPerl_Party_Events.PARTY_MEMBER_DISABLE	= XPerl_Party_Events.PARTY_MEMBER_ENABLE

-- UNIT_MAXHEALTH
function XPerl_Party_Events:UNIT_MAXHEALTH()
	XPerl_Party_UpdateHealth(self)
	XPerl_Party_UpdateLevel(self)	-- Level not available until we've received maxhealth
	XPerl_Party_UpdateClass(self)
end

-- PLAYER_ENTERING_WORLD
function XPerl_Party_Events:PLAYER_ENTERING_WORLD()
	XPerl_ProtectedCall(XPerl_Party_SetInitialAttributes)
end

-- RAID_ROSTER_UPDATED
function XPerl_Party_Events:RAID_ROSTER_UPDATED()
	checkRaidNextUpdate = 3
	CheckRaid()
end

XPerl_Party_Events.PARTY_MEMBERS_CHANGED = XPerl_Party_Events.RAID_ROSTER_UPDATED

-- UNIT_MODEL_CHANGED
function XPerl_Party_Events:UNIT_MODEL_CHANGED()
	XPerl_Party_UpdatePortrait(self)
end
--XPerl_Party_Events.UNIT_PORTRAIT_UPDATE = XPerl_Party_Events.UNIT_MODEL_CHANGED

-- UNIT_MANA
function XPerl_Party_Events:UNIT_MANA()
	XPerl_Party_UpdateMana(self)
end

XPerl_Party_Events.UNIT_MAXMANA		= XPerl_Party_Events.UNIT_MANA
XPerl_Party_Events.UNIT_RAGE		= XPerl_Party_Events.UNIT_MANA
XPerl_Party_Events.UNIT_MAXRAGE		= XPerl_Party_Events.UNIT_MANA
XPerl_Party_Events.UNIT_ENERGY		= XPerl_Party_Events.UNIT_MANA
XPerl_Party_Events.UNIT_MAXENERGY	= XPerl_Party_Events.UNIT_MANA

-- UNIT_DISPLAYPOWER
function XPerl_Party_Events:UNIT_DISPLAYPOWER()
	XPerl_Party_UpdateManaType(self)
	XPerl_Party_UpdateMana(self)
end

-- PLAYER_FLAGS_CHANGED()
function XPerl_Party_Events:PLAYER_FLAGS_CHANGED(unit)
	local f = PartyFrames[unit]
	if (f) then
		XPerl_Party_UpdatePlayerFlags(f)
	end
end

-- UNIT_NAME_UPDATE
function XPerl_Party_Events:UNIT_NAME_UPDATE()
	XPerl_Party_UpdateName(self)
end

-- UNIT_LEVEL
function XPerl_Party_Events:UNIT_LEVEL()
	XPerl_Party_UpdateLevel(self)
end

-- UNIT_AURA
function XPerl_Party_Events:UNIT_AURA()
	XPerl_Party_Buff_UpdateAll(self)
end

-- UNIT_FACTION
function XPerl_Party_Events:UNIT_FACTION(unit)
	XPerl_Party_UpdateName(self)
        XPerl_Party_UpdateCombat(self)
	XPerl_Party_Update_Control(self)
	XPerl_Party_UpdatePVP(self)
end

XPerl_Party_Events.UNIT_FLAGS = XPerl_Party_Events.UNIT_FACTION
XPerl_Party_Events.UNIT_DYNAMIC_FLAGS = XPerl_Party_Events.UNIT_FACTION

function XPerl_Party_Events:UNIT_TARGET()
	XPerl_Party_UpdateTarget(self)
end

---- Moving stuff ----
-- XPerl_Party_GetGap
function XPerl_Party_GetGap()
	--return floor(floor((XPerl_party1:GetBottom() - XPerl_party2:GetTop() + 0.01) * 100) / 100)
	return XPerlConfig.PartySpacing
end

-- XPerl_Party_SetGap
function XPerl_Party_SetGap(newGap)
	if (type(newGap) ~= "number") then
		return
	end
	XPerlConfig.PartySpacing = newGap
	XPerl_Party_SetMainAttributes()
	XPerl_Party_Virtual(true)
end

-- CalcWidth
local function CalcWidth(self)
	--return XPerl_party1Highlight:GetWidth()

	local w = self.statsFrame:GetWidth()

	if (XPerlConfig.ShowPartyPortrait == 1) then
		w = w + self.portraitFrame:GetWidth() - 2

		if (XPerlConfig.ShowPartyLevel == 1 or XPerlConfig.ShowPartyClassIcon == 1) then
			w = w + self.levelFrame:GetWidth() - 2
		end
	else
		w = w + self.levelFrame:GetWidth() - 2
	end

	return w
end

-- CalcHeight
local function CalcHeight(self)
	--return XPerl_party1Highlight:GetHeight()

	if (XPerlConfig.ShowPartyPortrait == 1) then
		return self.portraitFrame:GetHeight()
	end

	local h = self.statsFrame:GetHeight()

	if (XPerlConfig.ShowPartyNames == 1) then
		h = h + self.nameFrame:GetHeight() - 2
	end

	return h
end

-- XPerl_Party_Set_Bits
function XPerl_Party_Set_Bits1(self)
	local name = self:GetName()
	local levelFrame = self.levelFrame
	local nameFrame = self.nameFrame
	local statsFrame = self.statsFrame
	local portraitFrame = self.portraitFrame
	local clickFrame = getglobal(name.."Highlight")
	local classTexture = self.levelFrame.classTexture
	local levelText = self.levelFrame.text

	portraitFrame:ClearAllPoints()
	nameFrame:ClearAllPoints()
	levelFrame:ClearAllPoints()
	statsFrame:ClearAllPoints()
	classTexture:ClearAllPoints()
	levelText:ClearAllPoints()
	clickFrame:ClearAllPoints()

	if (XPerlConfig.ShowPartyPortrait == 0) then
		portraitFrame:Hide()

		levelFrame:SetWidth(30)
		levelFrame:SetHeight(41)

		nameFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
		levelFrame:SetPoint("TOPLEFT", nameFrame, "BOTTOMLEFT", 0, 3)
		statsFrame:SetPoint("TOPLEFT", levelFrame, "TOPRIGHT", -2, 0)

		levelText:SetPoint("BOTTOM", levelFrame, "BOTTOM", 0, 4)
		classTexture:SetPoint("TOPLEFT", levelFrame, "TOPLEFT", 5, -5)

		self.buffFrame:SetPoint("TOPLEFT", statsFrame, "BOTTOMLEFT", 5, 0)

		clickFrame:SetPoint("TOPLEFT", nameFrame, "TOPLEFT", 0, 0)
		clickFrame:SetPoint("BOTTOMRIGHT", statsFrame, "BOTTOMRIGHT", 0, 0)
	else
		portraitFrame:Show()

		levelFrame:SetWidth(27)
		levelFrame:SetHeight(22)

		levelFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 0, 0)
		portraitFrame:SetPoint("TOPLEFT", levelFrame, "TOPRIGHT", -2, 0)
		nameFrame:SetPoint("TOPLEFT", portraitFrame, "TOPRIGHT", -2, 0)
		statsFrame:SetPoint("TOPLEFT", nameFrame, "BOTTOMLEFT", 0, 3)

		levelText:SetPoint("CENTER", 0, 0)
		classTexture:SetPoint("BOTTOMRIGHT", portraitFrame, "BOTTOMLEFT", 0, 3)

		self.buffFrame:SetPoint("TOPLEFT", portraitFrame, "BOTTOMLEFT", 5, 0)

		clickFrame:SetPoint("TOPLEFT", levelFrame, "TOPLEFT", 0, 0)
		clickFrame:SetPoint("BOTTOMRIGHT", statsFrame, "BOTTOMRIGHT", 0, 0)
	end

	if (XPerlConfig.ShowPartyLevel==0) then
		self.levelFrame.text:Hide()
	else
		self.levelFrame.text:Show()
		levelFrame:Show()

		if (XPerlConfig.ShowPartyPortrait == 1) then
			levelFrame:SetWidth(27)
		else
			levelFrame:SetWidth(30)
		end
	end

	if (XPerlConfig.ShowPartyClassIcon==0) then
		classTexture:Hide()

		if (XPerlConfig.ShowPartyLevel == 0) then
			levelFrame:SetWidth(2)
			levelFrame:Hide()
		end
	else
		classTexture:Show()
		levelFrame:Show()

		if (XPerlConfig.ShowPartyPortrait == 1) then
			levelFrame:SetWidth(27)
		else
			levelFrame:SetWidth(30)
		end
	end

	local width = (36 * XPerlConfig.ShowPartyPercent) + 106;	-- 136 enabled, 106 disabled

	statsFrame:SetWidth(width)
	if (XPerlConfig.ShowPartyPercent == 0) then
		self.statsFrame.healthBar.percent:Hide()
		self.statsFrame.manaBar.percent:Hide()
	else
		self.statsFrame.healthBar.percent:Show()
		self.statsFrame.manaBar.percent:Show()
	end

	local height = (XPerlConfig.ShowPartyNames * 22) + 2;		-- 24 when enabled, 2 when disabled

	self.targetFrame:ClearAllPoints()
	nameFrame:SetHeight(height)

	if (XPerlConfig.ShowPartyNames==0) then
		nameFrame:Hide()
		self.targetFrame:SetPoint("TOPLEFT", statsFrame, "BOTTOMLEFT", 0, 2)
	else
		nameFrame:Show()
		self.targetFrame:SetPoint("BOTTOMLEFT", nameFrame, "BOTTOMRIGHT", -2, 2)
	end
	self.targetFrame:SetWidth(XPerlConfig.PartyTargetSize)

	XPerlConfig.PartyBuffSize = tonumber(XPerlConfig.PartyBuffSize) or 20
	XPerl_SetBuffSize(name.."_", XPerlConfig.PartyBuffSize, XPerlConfig.PartyBuffSize * 1.4)

	self:SetWidth(CalcWidth(self))

	if (XPerlConfig.ShowPartyTarget == 1) then
		RegisterUnitWatch(self.targetFrame)
	else
		UnregisterUnitWatch(self.targetFrame)
		self.targetFrame:Hide()
	end

	XPerl_StatsFrameSetup(self)

	XPerl_Party_UpdateDisplay(self)
end

-- XPerl_Party_SetInitialAttributes()
function XPerl_Party_SetInitialAttributes()
	XPerl_Party_SecureHeader:Hide()
	XPerl_Party_SecureHeader:SetAttribute("minWidth", 80)
	XPerl_Party_SecureHeader:SetAttribute("minHeight", 10)
	XPerl_Party_SecureHeader:SetAttribute("xOffset", 0)
	XPerl_Party_SecureHeader:SetAttribute("template", "XPerl_Party_FrameTemplate")
	XPerl_Party_SecureHeader:SetAttribute("templateType", "Button")
	XPerl_Party_SetMainAttributes()
	CheckRaid()
end

-- XPerl_Party_SetMainAttributes
function XPerl_Party_SetMainAttributes()

	XPerl_Party_Anchor:StopMovingOrSizing()

	XPerl_Party_SecureHeader:ClearAllPoints()
	if (XPerlConfig.PartyAnchor == "TOP") then
		XPerl_Party_SecureHeader:SetPoint("TOPLEFT", XPerl_Party_Anchor, "TOPLEFT", 0, 0)
        	XPerl_Party_SecureHeader:SetAttribute("xOffset", 0)
		XPerl_Party_SecureHeader:SetAttribute("yOffset", -XPerlConfig.PartySpacing)
	elseif (XPerlConfig.PartyAnchor == "LEFT") then
		XPerl_Party_SecureHeader:SetPoint("TOPLEFT", XPerl_Party_Anchor, "TOPLEFT", 0, 0)
		XPerl_Party_SecureHeader:SetAttribute("xOffset", XPerlConfig.PartySpacing)
        	XPerl_Party_SecureHeader:SetAttribute("yOffset", 0)
	elseif (XPerlConfig.PartyAnchor == "BOTTOM") then
		XPerl_Party_SecureHeader:SetPoint("BOTTOMLEFT", XPerl_Party_Anchor, "BOTTOMLEFT", 0, 0)
        	XPerl_Party_SecureHeader:SetAttribute("xOffset", 0)
		XPerl_Party_SecureHeader:SetAttribute("yOffset", XPerlConfig.PartySpacing)
	elseif (XPerlConfig.PartyAnchor == "RIGHT") then
		XPerl_Party_SecureHeader:SetPoint("BOTTOMRIGHT", XPerl_Party_Anchor, "BOTTOMRIGHT", 0, 0)
		XPerl_Party_SecureHeader:SetAttribute("xOffset", -XPerlConfig.PartySpacing)
        	XPerl_Party_SecureHeader:SetAttribute("yOffset", 0)
	end

        XPerl_Party_SecureHeader:SetAttribute("point", XPerlConfig.PartyAnchor or "TOP")
end

-- XPerl_Party_Virtual
function XPerl_Party_Virtual(on)

	local conf = XPerlConfig
	if (on) then
		local w = CalcWidth(XPerl_party1)
		local h = CalcHeight(XPerl_party1)

		XPerl_Party_AnchorVirtual:ClearAllPoints()
		if (conf.PartyAnchor == "TOP") then
			XPerl_Party_AnchorVirtual:SetPoint("TOPLEFT", XPerl_Party_Anchor, "TOPLEFT", 0, 0)
			XPerl_Party_AnchorVirtual:SetHeight((h * 4) + (conf.PartySpacing * 3))
			XPerl_Party_AnchorVirtual:SetWidth(w)

		elseif (conf.PartyAnchor == "LEFT") then
			XPerl_Party_AnchorVirtual:SetPoint("TOPLEFT", XPerl_Party_Anchor, "TOPLEFT", 0, 0)
			XPerl_Party_AnchorVirtual:SetHeight(h)
			XPerl_Party_AnchorVirtual:SetWidth(w * 4 + (conf.PartySpacing * 3))

		elseif (conf.PartyAnchor == "BOTTOM") then
			XPerl_Party_AnchorVirtual:SetPoint("BOTTOMLEFT", XPerl_Party_Anchor, "BOTTOMLEFT", 0, 0)
			XPerl_Party_AnchorVirtual:SetHeight((h * 4) + (conf.PartySpacing * 3))
			XPerl_Party_AnchorVirtual:SetWidth(w)

		elseif (conf.PartyAnchor == "RIGHT") then
			XPerl_Party_AnchorVirtual:SetPoint("TOPRIGHT", XPerl_Party_Anchor, "TOPRIGHT", 0, 0)
			XPerl_Party_AnchorVirtual:SetHeight(h)
			XPerl_Party_AnchorVirtual:SetWidth(w * 4 + (conf.PartySpacing * 3))
		end

		XPerl_Party_AnchorVirtual:SetBackdropColor(conf.BackColour.r, conf.BackColour.g, conf.BackColour.b, conf.BackColour.a)
		XPerl_Party_AnchorVirtual:SetBackdropBorderColor(conf.BorderColour.r, conf.BorderColour.g, conf.BorderColour.b, 1)
		XPerl_Party_AnchorVirtual:Lower()
		XPerl_Party_AnchorVirtual:Show()
	else
		XPerl_Party_AnchorVirtual:Hide()
	end
end

-- XPerl_Party_Set_Bits
function XPerl_Party_Set_Bits()
	if (InCombatLockdown()) then
		tinsert(XPerl_OutOfCombatQueue, XPerl_Party_Set_Bits)
		return
	end

	for k,v in pairs(PartyFrames) do
		XPerl_Party_Set_Bits1(v)
	end

	XPerl_Party_SecureHeader:Hide()
	XPerl_Party_SetMainAttributes()
	CheckRaid()

	if (XPerl_Party_AnchorVirtual:IsShown()) then
		XPerl_Party_Virtual(true)
	end

	XPerl_Party_Anchor:SetScale(XPerlConfig.Scale_PartyFrame)
end
